"""Top-level package for Onacol."""

__author__ = """Josef Nevrly"""
__email__ = 'josef.nevrly@gmail.com'
__version__ = '0.2.3'

from .onacol import ConfigManager, OnacolException, ConfigValidationError
